import types
from lab3.constants import NULL, TRUE, FALSE, QUOTATION_MARK, MINUS, EXTRA_CHARACTERS_IN_LIST, \
    EXTRA_CHARACTERS_IN_DICT


def deserialize(s, index=0):
    if s[index:index + 4] == NULL:
        index += 4

        return None, index

    elif s[index:index + 4] == TRUE:
        index += 4

        return True, index

    elif s[index:index + 5] == FALSE:
        index += 5

        return False, index

    elif s[index].isdigit() or s[index] == MINUS:
        result, index = deserialize_number(s, index)

        return result, index

    elif s[index] == QUOTATION_MARK:
        result, index = deserialize_string(s, index + 1)

        return result, index

    elif s[index] == '[':
        result, index = deserialize_list(s, index)

        return result, index

    elif s[index] == '{':
        result, index = deserialize_dict(s, index)

        return result, index

    else:
        raise TypeError


def deserialize_class(cls):
    if cls['class_type']['__name__'] == 'object':
        return object

    else:
        result = type(cls['class_type']["__name__"], tuple(cls['class_type']["__bases__"]),
                      cls['class_type']["__code__"])

        return result


def deserialize_dict(s, index):
    result = {}
    in_key = True
    now_key = None
    index += 1

    while s[index] != '}':
        if s[index] in EXTRA_CHARACTERS_IN_DICT:
            index += 1

        elif s[index] == QUOTATION_MARK and in_key:
            now_key, index = deserialize(s, index)
            result[now_key] = None
            in_key = False

        else:
            value, index = deserialize(s, index)
            result[now_key] = value
            in_key = True

    if 'function_type' in result:
        result = deserialize_function(result)

    elif 'class_type' in result:
        result = deserialize_class(result)

    elif 'type_code' in result:
        result = deserialize_code(result)
    elif 'instance_type' in result:
        result = deserialize_instance(result)
    return result, index + 1


def deserialize_instance(obj):
    def __init__(self):
        pass

    cls = obj['instance_type']['class']
    temp = cls.__init__
    cls.__init__ = __init__
    result = obj['instance_type']['class']()
    result.__dict__ = obj['instance_type']['dict']
    result.__init__ = temp
    result.__class__.__init__ = temp
    return result


def get_code(obj):
    result = (obj['co_argcount'],
              obj['co_posonlyargcount'],
              obj['co_kwonlyargcount'],
              obj['co_nlocals'],
              obj['co_stacksize'],
              obj['co_flags'],
              bytes(obj['co_code']),
              tuple(obj['co_consts']),
              tuple(obj['co_names']),
              tuple(obj['co_varnames']),
              obj['co_filename'],
              obj['co_name'],
              obj['co_firstlineno'],
              bytes(obj['co_lnotab']),
              tuple(obj['co_freevars']),
              tuple(obj['co_cellvars']))

    return result


def deserialize_code(obj: dict):
    result = types.CodeType(*get_code(obj['type_code']))

    return result


def deserialize_function(obj: dict):
    my_code = get_code(obj['function_type']['__code__']['code_type'])
    result = types.FunctionType(types.CodeType(*my_code), obj['function_type']['__globals__'],
                                obj['function_type']['__name__'])

    for key in obj['function_type']['__globals__']:

        try:
            result.__globals__[key] = __import__(obj['function_type']['__globals__'][key])
        except:
            pass

    result.__globals__.update({result.__name__: result})
    result.__globals__["__builtins__"] = __import__("builtins")

    return result




def deserialize_list(s, index):
    result = []
    index += 1

    while s[index] != ']':
        if s[index] in EXTRA_CHARACTERS_IN_LIST:
            index += 1

        else:
            obj, index = deserialize(s, index)

            result.append(obj)

    return result, index + 1


def deserialize_number(s: str, index):
    end = index

    while end != len(s) and (s[end].isdigit() or s[end] == '.' or s[end] == '-'):
        end += 1

    try:
        result = int(s[index:end])

    except ValueError:
        try:
            result = float(s[index:end])

        except ValueError:
            raise ValueError

    return result, end + 1


def deserialize_string(s, index):
    end = index

    while s[end] != QUOTATION_MARK:
        end += 1

    return s[index:end], end + 1



import yaml
from lab3.SerializerYaml import serialize
from lab3.DeserializerYaml import deserialize


class Yaml:
    @staticmethod
    def dump(obj, f):
        data = serialize(obj)
        with open(f, 'w') as file:
            yaml.dump(data, file)

    @staticmethod
    def dumps(obj):
        data = serialize(obj)
        result = yaml.dump(data)
        return result

    @staticmethod
    def load(f):
        with open(f, 'r') as file:
            data = yaml.unsafe_load(file)
        result = deserialize(data)
        return result

    @staticmethod
    def loads(s):
        data = yaml.unsafe_load(s)
        result = deserialize(data)
        return result



import types
import inspect
import lab3.DeserializerJson
from lab3.constants import NULL, TRUE, FALSE, QUOTATION_MARK, EXTRA_ATTRIBUTE_CLASS_CODE


def serialize(obj, indent, new_indent=0):
    if obj is None:
        return NULL

    elif obj is True:
        return TRUE

    elif obj is False:
        return FALSE

    elif isinstance(obj, (int, float)):
        return str(obj)

    elif isinstance(obj, str):
        if obj[0] == QUOTATION_MARK:
            result = obj
        else:
            result = QUOTATION_MARK + obj + QUOTATION_MARK
        return result

    elif isinstance(obj, dict):
        return serialize_dict(obj, indent, new_indent)

    elif isinstance(obj, list) or isinstance(obj, tuple) or isinstance(obj, set) or isinstance(obj, frozenset):
        return serialize_list(obj, indent, new_indent)

    elif inspect.isclass(obj):
        return serialize_class(obj, indent, new_indent)

    elif inspect.isfunction(obj):
        return serialize_function(obj, indent, new_indent)

    elif isinstance(obj, types.CodeType):
        return serialize_code(obj, indent, new_indent)

    elif inspect.ismethod(obj):
        return serialize_function(obj, indent, new_indent)

    else:
        try:
            return serialize_instance(obj, indent, new_indent)
        except TypeError:
            raise TypeError


def serialize_instance(obj, indent, new_indent):
    data = {
        'instance_type': {
            'class': obj.__class__,
            'dict': obj.__dict__,
        }
    }

    result = serialize_dict(data, indent, new_indent)

    return result


def serialize_class(obj, indent, new_indent):
    class_dict = {
        "class_type": {
            "__name__": get_name_class(obj, indent, new_indent),
            "__bases__": tuple(get_bases_class(obj, indent, new_indent)),
            "__code__": get_code_class(obj, indent, new_indent)
        }
    }

    result = serialize_dict(class_dict, indent, new_indent)

    return result


def get_name_class(obj, indent, new_indent):
    result = serialize(obj.__name__, indent, new_indent)

    return result


def get_bases_class(obj, indent, new_indent):
    data = serialize_list(obj.__bases__, indent, new_indent)
    result = lab3.DeserializerJson.deserialize_list(data, 0)[0]

    return result


def serialize_dict_class(obj, indent, new_indent=0):
    if len(obj) == 0:
        return '{}'

    else:
        result = '{\n'
        new_indent += indent

        for key in list(obj)[:len(obj) - 1]:

            if key in EXTRA_ATTRIBUTE_CLASS_CODE:
                continue

            result += ' ' * new_indent + QUOTATION_MARK + str(key) + QUOTATION_MARK + ': ' \
                      + serialize(obj[key], indent, new_indent) + ',\n'

        if list(obj)[len(obj) - 1] not in EXTRA_ATTRIBUTE_CLASS_CODE:
            result += ' ' * new_indent + QUOTATION_MARK + str(list(obj)[len(obj) - 1]) \
                      + QUOTATION_MARK + ': ' + serialize(obj[list(obj)[len(obj) - 1]], indent, new_indent) + '\n'

        result += ' ' * (new_indent - indent) + '}'

    return result


def get_code_class(obj, indent, new_indent):
    if obj.__name__ != 'object':
        data = serialize_dict_class(dict(obj.__dict__), indent, new_indent)
        result = lab3.DeserializerJson.deserialize_dict(data, 0)[0]

    else:
        result = {}

    return result


def get_globals(obj):
    result = {}

    for key in obj.__globals__:

        if isinstance(obj.__globals__[key], types.ModuleType):
            result[key] = obj.__globals__[key].__name__

        elif key in obj.__code__.co_names and obj.__name__ != key:
            result[key] = obj.__globals__[key]

    return result


def serialize_code(obj: types.CodeType, indent, new_indent):
    code_dict = {
        "type_code": {
            "co_argcount": obj.co_argcount,
            "co_posonlyargcount": obj.co_posonlyargcount,
            "co_kwonlyargcount": obj.co_kwonlyargcount,
            "co_nlocals": obj.co_nlocals,
            "co_stacksize": obj.co_stacksize,
            "co_flags": obj.co_flags,
            "co_code": list(obj.co_code),
            "co_consts": obj.co_consts,
            "co_names": obj.co_names,
            "co_varnames": obj.co_varnames,
            "co_filename": obj.co_filename,
            "co_name": obj.co_name,
            "co_firstlineno": obj.co_firstlineno,
            "co_lnotab": list(obj.co_lnotab),
            "co_freevars": obj.co_freevars,
            "co_cellvars": obj.co_cellvars
        }
    }

    result = serialize_dict(code_dict, indent, new_indent)

    return result


def serialize_function(obj: types.FunctionType, indent, new_indent):
    glob = get_globals(obj)

    function_dict = {
        "function_type": {
            "__globals__": glob,
            "__name__": obj.__name__,
            "__code__": {
                "code_type": {
                    "co_argcount": obj.__code__.co_argcount,
                    "co_posonlyargcount": obj.__code__.co_posonlyargcount,
                    "co_kwonlyargcount": obj.__code__.co_kwonlyargcount,
                    "co_nlocals": obj.__code__.co_nlocals,
                    "co_stacksize": obj.__code__.co_stacksize,
                    "co_flags": obj.__code__.co_flags,
                    "co_code": list(obj.__code__.co_code),
                    "co_consts": obj.__code__.co_consts,
                    "co_names": obj.__code__.co_names,
                    "co_varnames": obj.__code__.co_varnames,
                    "co_filename": obj.__code__.co_filename,
                    "co_name": obj.__code__.co_name,
                    "co_firstlineno": obj.__code__.co_firstlineno,
                    "co_lnotab": list(obj.__code__.co_lnotab),
                    "co_freevars": obj.__code__.co_freevars,
                    "co_cellvars": obj.__code__.co_cellvars
                }
            }
        }
    }

    result = serialize_dict(function_dict, indent, new_indent)

    return result


def serialize_dict(obj, indent, new_indent=0):
    if len(obj) == 0:
        return '{}'

    else:
        result = '{\n'
        new_indent += indent

        for key in list(obj)[:len(obj) - 1]:
            result += ' ' * new_indent + QUOTATION_MARK + str(key) + QUOTATION_MARK + ': ' \
                      + serialize(obj[key], indent, new_indent) + ',\n'

        result += ' ' * new_indent + QUOTATION_MARK + str(list(obj)[len(obj) - 1]) + QUOTATION_MARK + ': ' + \
                  serialize(obj[list(obj)[len(obj) - 1]], indent, new_indent) + '\n'
        result += ' ' * (new_indent - indent) + '}'

    return result


def serialize_list(obj, indent, new_indent=0):
    if len(obj) == 0:
        return '[]'

    else:
        result = '[\n'
        new_indent += indent

        for el in obj[:len(obj) - 1]:
            result += ' ' * new_indent + serialize(el, indent, new_indent) + ',\n'

        result += ' ' * new_indent + serialize(obj[len(obj) - 1], indent, new_indent) + '\n'
        result += ' ' * (new_indent - indent) + ']'

    return result



JSON = 'json'
YAML = 'yaml'
TOML = 'toml'
NULL = 'null'
TRUE = 'true'
FALSE = 'false'
QUOTATION_MARK = '"'
MINUS = '-'
EXTRA_CHARACTERS_IN_LIST = [',', '\n', ' ']
EXTRA_CHARACTERS_IN_DICT = [',', '\n', ' ', ':']
EXTRA_ATTRIBUTE_CLASS_CODE=['__module__','__dict__','__weakref__']



from lab3.constants import JSON, YAML, TOML
from lab3 import Json
from lab3 import Toml
from lab3 import Yaml


class Serializer:

    @staticmethod
    def create_serializer(t):
        if t == JSON:
            return Json

        elif t == YAML:
            return Yaml

        elif t == TOML:
            return Toml

        else:
            raise ValueError



import types


def deserialize(obj):
    if isinstance(obj, (int, float, bool)):
        return obj
    if isinstance(obj, str):
        if str == 'None':
            return None
        else:
            return obj
    elif isinstance(obj, list):
        return deserialize_list(obj)
    elif isinstance(obj, tuple):
        return deserialize_tuple(obj)
    elif isinstance(obj, dict):
        if 'function_type' in obj:
            return deserialize_function(obj)
        elif 'type_code' in obj:
            return deserialize_code(obj)
        elif 'class_type' in obj:
            return deserialize_class(obj)
        elif 'instance_type' in obj:
            return deserialize_instance(obj)
        else:
            return deserialize_dict(obj)
    else:
        raise TypeError


def deserialize_class(cls):
    if cls['class_type']['__name__'] == 'object':
        return object

    else:
        x = deserialize(cls['class_type']["__bases__"])
        y = deserialize(cls['class_type']["__code__"])
        result = type(cls['class_type']["__name__"], tuple(deserialize(cls['class_type']["__bases__"])),
                      deserialize(cls['class_type']["__code__"]))

        return result


def deserialize_dict(obj):
    result = {}

    for key in obj:
        if obj[key] == "None":
            result[key] = None
        else:
            result[key] = deserialize(obj[key])

    return result


def deserialize_instance(obj):
    def __init__(self):
        pass

    cls = obj['instance_type']['class']
    temp = cls.__init__
    cls.__init__ = __init__
    result = obj['instance_type']['class']()
    result.__dict__ = obj['instance_type']['dict']
    result.__init__ = temp
    result.__class__.__init__ = temp
    return result


def get_code(obj):
    result = (obj['co_argcount'],
              obj['co_posonlyargcount'],
              obj['co_kwonlyargcount'],
              obj['co_nlocals'],
              obj['co_stacksize'],
              obj['co_flags'],
              bytes(obj['co_code']),
              deserialize(tuple(obj['co_consts'])),
              tuple(obj['co_names']),
              tuple(obj['co_varnames']),
              obj['co_filename'],
              obj['co_name'],
              obj['co_firstlineno'],
              bytes(obj['co_lnotab']),
              tuple(obj['co_freevars']),
              tuple(obj['co_cellvars']))

    return result


def deserialize_code(obj: dict):
    result = types.CodeType(*get_code(obj['type_code']))

    return result


def deserialize_function(obj: dict):
    my_code = get_code(obj['function_type']['__code__']['code_type'])
    result = types.FunctionType(types.CodeType(*my_code), deserialize(obj['function_type']['__globals__']),
                                obj['function_type']['__name__'])

    for key in obj['function_type']['__globals__']:

        try:
            result.__globals__[key] = __import__(obj['function_type']['__globals__'][key])
        except:
            pass

    result.__globals__.update({result.__name__: result})
    result.__globals__["__builtins__"] = __import__("builtins")

    return result


def deserialize_list(obj):
    result = []

    for el in obj:
        if el == "None":
            result.append(None)
        else:
            result.append(deserialize(el))

    return result


def deserialize_tuple(obj):
    result = tuple(deserialize_list(obj))

    return result


def deserialize_instance(obj):
    obj['instance_type']['__class__'] = deserialize(obj['instance_type']['__class__'])

    def __init__(self):
        pass

    cls = obj['instance_type']['__class__']
    temp = cls.__init__
    cls.__init__ = __init__
    result = obj['instance_type']['__class__']()
    result.__dict__ = obj['instance_type']['__dict__']
    result.__init__ = temp
    result.__class__.__init__ = temp
    return result



import lab3.DeserializerJson
from lab3 import SerializerJson


class Json:
    @staticmethod
    def dump(obj, fp, indent=0):
        result = Json.dumps(obj, indent)

        with open(fp, 'w') as file:
            file.write(result)

    @staticmethod
    def dumps(obj, indent=0):
        result = SerializerJson.serialize(obj, indent)

        return result

    @staticmethod
    def load(fp):
        with open(fp, 'r') as file:
            data = file.read()

        result = Json.loads(data)

        return result

    @staticmethod
    def loads(s):
        result = lab3.DeserializerJson.deserialize(s)[0]

        return result



import pytomlpp
import lab3.SerializerYaml
import lab3.DeserializerYaml

class Toml:
    @staticmethod
    def dump(obj, f):
        data = lab3.SerializerYaml.serialize(obj)
        with open(f, 'w') as file:
            pytomlpp.dump(data, file)

    @staticmethod
    def dumps(obj):
        data = lab3.SerializerYaml.serialize(obj)
        return pytomlpp.dumps(data)

    @staticmethod
    def load(f):
        with open(f, 'r+') as file:
            data = pytomlpp.load(file)
        return lab3.DeserializerYaml.deserialize(data)

    @staticmethod
    def loads(s):
        data = pytomlpp.loads(s)
        result = lab3.DeserializerYaml.deserialize(data)

        return result



from setuptools import setup

setup(
    name='twoserializer',
    version="2.0.4",
    description="serializer of few formats",
    author="Eb1geil",
    author_email='ksanvei03@gmail.com',
    url='https://github.com/Eb1geil/twoserializer',
    install_requires=["pytomlpp", 'pyyaml'],
    packages=['lab3'],
    test_suite='tests/',
)
